/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include "config_xor.h"

#ifdef HAVE_BROKEN_INCLUDES
#define _ANSI_C_SOURCE
#define _POSIX_SOURCE
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>


#include "samhain.h"
#include "sh_error.h"
#include "sh_unix.h"
#include "sh_utils.h"
#include "sh_files.h"
#include "sh_mail.h"
#include "sh_tiger.h"
#include "sh_forward.h"
#include "sh_modules.h"
#ifdef WITH_EXTERNAL
#include "sh_extern.h"
#endif


typedef enum {
  SH_SECTION_NONE,
  SH_SECTION_LOG,
  SH_SECTION_MISC,
  SH_SECTION_ATTRIBUTES,
  SH_SECTION_READONLY,
  SH_SECTION_LOGFILES,
  SH_SECTION_LOGGROW,
  SH_SECTION_NOIGNORE,
  SH_SECTION_ALLIGNORE,
#ifdef WITH_EXTERNAL
  SH_SECTION_EXTERNAL,
#endif
#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
  SH_SECTION_OTHER,
#endif
#ifdef SH_WITH_SERVER
  SH_SECTION_CLIENTS,
#endif
  SH_SECTION_THRESHOLD
} ShSectionType;
    
static void sh_readconfig_line (char * line);

static ShSectionType read_mode = SH_SECTION_NONE;

static int conf_line = 0;

/* read the configuration file
 */
int sh_readconf_read ()
{
#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
  /* this is for modules */
  int    modnum;
#endif

  SL_TICKET    fd = -1;
  char * tmp;
  char * lptr;

  char   line_in[512+2];
  char * line;

  char   myident[3*SH_MINIBUF+3];

  /* this is for nested conditionals
   */
  int    some_other_host[16]   = { 0 };
  int    some_other_system[16] = { 0 };
  int    seen_host   = 0;
  int    seen_system = 0;
  int    host_int    = 0;
  int    sys_int     = 0;

  int    invert = 0;
  int    length = sl_strlen(sh.localhost);

  int    local_file = 1;
  char   local_flag = 'R';

  /* open config file, exit on failure
   */
#if defined(SH_WITH_CLIENT)
  if (0 == sl_strcmp(file_path('C', 'R'), _("REQ_FROM_SERVER")))
    {
      sh_error_handle (SH_ERR_INFO,
		       _("Downloading configuration file."));
      fd = sh_forward_req_file(_("CONF"));

      if (!SL_ISERROR(fd))
	local_file = 0;
      else if (sh.checkSum != SH_CHECK_INIT)
	exit (EXIT_FAILURE);
      else
	{
	  sh_error_handle (SH_ERR_INFO,
			   _("No file from server, trying local file."));
	  local_file = 1;
	  local_flag = 'I';
	}
    }
#endif

  /* Use a local configuration file.
   */
  if (local_file == 1)
    {
      if (0 != tf_trust_check (file_path('C', local_flag), SL_YESPRIV)
	  ||
	  SL_ISERROR(fd = sl_open_read(file_path('C',local_flag),SL_YESPRIV)))
	{
	  sh_error_handle (SH_ERR_FATAL,
			   _("PANIC -- cannot open configuration file %s"),
			   file_path('C', local_flag));
	  exit (EXIT_FAILURE);
	}

    }

  tiger_fd = fd;
  sl_strlcpy(sh.cfg_hash, sh_tiger_hash(file_path('C',local_flag),TIGER_FD, 0),
	     KEY_LEN+1);

  sl_rewind (fd);

  sprintf(myident, _("%s:%s:%s"),                  /* known to fit  */
	  sh.sh_sysname, sh.sh_release, sh.sh_machine);


#if defined(SH_STEALTH) && defined(SH_STEALTH_MICRO)
  while ( sh_unix_getline (fd, line_in, 512) > 0) {
#elif !defined(SH_STEALTH)
  while ( sh_unix_getline (fd, line_in, 512) > 0) {
#else
  while ( sh_unix_getline_stealth (fd, line_in, 512) > 0) {
#endif

    ++conf_line;

    line = &(line_in[0]);

    /* skip leading white space
     */
    while (isspace(*line))
      ++line;


    /* skip header etc. 
     */
    if (line[0] == '#' || line[0] == '\0' || line[0] == ';' || line[0] == '/')
      continue; 
  
    /* clip off trailing white space                 
     */
    tmp = line + sl_strlen( line ); --tmp;
    while( isspace((int) *tmp ) && tmp >= line ) *tmp-- = '\0';

    /* -------  starts a section  ------------  */
    
    if (line[0] == '['                       && 
	some_other_host[seen_host] == 0      &&
	some_other_system[seen_system] == 0) 
      { 
	read_mode = SH_SECTION_NONE;

	if      (sl_strncmp (line,  _("[ReadOnly]"), 
			  sizeof("[ReadOnly]")-1) == 0) 
	  read_mode = SH_SECTION_READONLY;
	else if (sl_strncmp (line,  _("[Attributes]"), 
			  sizeof("[Attributes]")-1) == 0)
	  read_mode = SH_SECTION_ATTRIBUTES;
	else if (sl_strncmp (line,  _("[LogFiles]"), 
			  sizeof("[LogFiles]")-1) == 0) 
	  read_mode = SH_SECTION_LOGFILES;
	else if (sl_strncmp (line,  _("[GrowingLogFiles]"), 
			  sizeof("[GrowingLogFiles]")-1) == 0) 
	  read_mode = SH_SECTION_LOGGROW;
	else if (sl_strncmp (line,  _("[IgnoreAll]"), 
			  sizeof("[IgnoreAll]")-1) == 0) 
	  read_mode = SH_SECTION_ALLIGNORE;
	else if (sl_strncmp (line,  _("[IgnoreNone]"), 
			  sizeof("[IgnoreNone]")-1) == 0) 
	  read_mode = SH_SECTION_NOIGNORE;
	else if (sl_strncmp (line,  _("[Log]"), 
			  sizeof("[Log]")-1) == 0) 
	  read_mode = SH_SECTION_LOG;
#ifdef WITH_EXTERNAL
	else if (sl_strncmp (line,  _("[External]"), 
			  sizeof("[External]")-1) == 0) 
	  read_mode = SH_SECTION_EXTERNAL;
#endif
	else if (sl_strncmp (line,  _("[Misc]"), 
			  sizeof("[Misc]")-1) == 0) 
	  read_mode = SH_SECTION_MISC;
	else if (sl_strncmp (line,  _("[EOF]"), 
			  5) == 0)
	  {
	    sl_close (fd);
	    return (0);
	  }
	else if (sl_strncmp (line,  _("[EventSeverity]"), 
			  sizeof("[EventSeverity]")-1) == 0) 
	  read_mode = SH_SECTION_THRESHOLD;

#ifdef SH_WITH_SERVER
	else if (sl_strncmp (line,  _("[Clients]"), 
			  sizeof("[Clients]")-1) == 0) 
	  read_mode = SH_SECTION_CLIENTS;
#endif

#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
	else
	  for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
	    {
	      if (0 == sl_strncmp (line,  _(modList[modnum].conf_section),
				sl_strlen(modList[modnum].conf_section)) )
		read_mode = SH_SECTION_OTHER;
	    }
#endif
	if (read_mode == SH_SECTION_NONE)
	  sh_error_handle (SH_ERR_ALL,
			   _("config file: unrecognized section in line %d"),
			   conf_line);
      } 

    /* ---  an @host directive -------------- */

    else if (line[0] == '@' || (line[0] == '!' && line[1] == '@')) 
      {
	if (line[0] == '!')
	  {
	    lptr   = &line[2];
	    invert = 1;
	  }
	else
	  {
	    lptr   = &line[1];
	    invert = 0;
	  }

	if (sl_strncmp (lptr, _("end"), 3) == 0)
	  {
	    if (0 == seen_host)
	      sh_error_handle (SH_ERR_WARN,
			       _("config file: unmatched @end in line %d"),
			       conf_line);

	    else if (host_int == 0)
	      {
		/* end  of an @host directive 
		 */
		some_other_host[seen_host] = 0;
		--seen_host;
		seen_host = (seen_host < 0) ? 0 : seen_host;
	      }

	    else 
	      {
		--host_int;
		host_int = (host_int < 0) ? 0 : host_int;
	      }
	  }
	else if (some_other_host[seen_host] == 0      &&
		 some_other_system[seen_system] == 0  && 
		 seen_host < 15)
	  {
	    if  (sl_strncmp (lptr,  sh.localhost, length) == 0
#ifdef HAVE_REGEX_H
		 || sh_util_regcmp (lptr, sh.localhost) == 0
#endif
		 )
	      {
		/* if match and '@',  set some_other_host = 0;
		 * if match and '!@', set some_other_host = 1;
		 */
		++seen_host;
		some_other_host[seen_host] = invert; 
	      }
	    else
	      {
		/* if no match and '@',  set some_other_host = 1;
		 * if no match and '!@', set some_other_host = 0;
		 */
		++seen_host;
		some_other_host[seen_host] = (invert == 0) ? 1 : 0;
	      }
	  }
	else
	  ++host_int;
      } 

    /* ---  an $system directive -------------- */

    else if (line[0] == '$' || (line[0] == '!' && line[1] == '$')) 
      {
	if (line[0] == '!')
	  {
	    lptr   = &line[2];
	    invert = 1;
	  }
	else
	  {
	    lptr   = &line[1];
	    invert = 0;
	  }

	if (sl_strncmp (lptr, _("end"), 3) == 0)
	  {
	    if (0 == seen_system)
	      sh_error_handle (SH_ERR_WARN, 
			       _("config file: unmatched $end in line %d"),
			       conf_line);
	    else if (sys_int == 0)
	      {
		/* end  of an $system directive 
		 */
		some_other_system[seen_system] = 0;
		--seen_system;
		seen_system = (seen_system < 0) ? 0 : seen_system;
	      }
	    else 
	      {
		--sys_int;
		sys_int = (sys_int < 0) ? 0 : sys_int;
	      }
	  }
	else if (some_other_host[seen_host] == 0      &&
		 some_other_system[seen_system] == 0  && 
		 seen_system < 15)
	  {
	    if  (sl_strncmp (lptr,  myident, sl_strlen(myident)) == 0
#ifdef HAVE_REGEX_H
		 || sh_util_regcmp (lptr, myident) == 0
#endif
		 )
	      {
		++seen_system;
		some_other_system[seen_system] = invert;
	      }
	    else
	      {
		++seen_system;
		some_other_system[seen_system] = (invert == 0) ? 1 : 0;
	      }
	  }
	else
	  ++sys_int;
      }

    /* ------  no new section -------------- */


    else if (some_other_host[seen_host] == 0          && 
	     some_other_system[seen_system] == 0      && 
	     read_mode != SH_SECTION_NONE) 
      { 
	sh_readconfig_line (line);
      }

  }
	   
  if (0 != seen_host)
    sh_error_handle (SH_ERR_WARN,
		     _("config file: missing @end."));

  if (0 != seen_system)
    sh_error_handle (SH_ERR_WARN,
		     _("config file: missing $end."));


  sl_close (fd);
  return (0);
}

typedef enum {
  SET_MAILTIME,
  SET_FILETIME 
} ShTimerItem;

    
int sh_readconf_setTime (char * str, ShTimerItem what)
{
  unsigned long i = atoi (str);

  if (i >= 0 && i < LONG_MAX) 
    {
      if      (what == SET_MAILTIME) 
	sh.mailTime.alarm_interval = i;
      else if (what == SET_FILETIME) 
	sh.fileCheck.alarm_interval  = i;
      return (0);
    } 
  else 
    {
      sh_error_handle (SH_ERR_ERR,
		       _("Invalid argument for maximum time interval (%s)"),
		       str);
      return (-1);
    }
}

#if defined(WITH_EXTERNAL)
typedef struct _cfg_options {
  char * optname;
  int (*func)(char * opt);
} cfg_options;

cfg_options ext_table[] = {
  { N_("OpenCommand"),  
    sh_ext_setcommand },
  { N_("SetCommandline"),  
    sh_ext_add_argv },
  { N_("SetChecksum"),  
    sh_ext_checksum },
  { N_("SetDefault"),  
    sh_ext_add_default },
  { N_("SetEnviron"),  
    sh_ext_add_environ },
  { N_("SetDeadtime"),  
    sh_ext_deadtime },
  { N_("SetType"),  
    sh_ext_type },
  { N_("SetCredentials"),  
    sh_ext_priv },

  { N_("SetFilterNot"),  
    sh_ext_add_not },
  { N_("SetFilterAnd"),  
    sh_ext_add_and },
  { N_("SetFilterOr"),  
    sh_ext_add_or },
  { NULL,
    NULL }
};

#endif



static void sh_readconfig_line (char * line)
{    
  char * c;
  char * tmp;
  int    i;
#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
  int    modnum, modkey;
#endif
  int    key = 0;

  static char  *ident[] = {
    N_("SeverityReadOnly"),
    N_("SeverityLogFiles"),
    N_("SeverityGrowingLogs"),
    N_("SeverityIgnoreNone"),
    N_("SeverityIgnoreAll"),
    N_("SeverityAttributes"),
    N_("SeverityDirs"),
    N_("SeverityFiles"),
    N_("SeverityNames"),
    NULL
  };

  static int      identnum[] = { 
    SH_ERR_T_RO,    
    SH_ERR_T_LOGS,  
    SH_ERR_T_GLOG,  
    SH_ERR_T_NOIG,  
    SH_ERR_T_ALLIG, 
    SH_ERR_T_ATTR, 
    SH_ERR_T_DIR,   
    SH_ERR_T_FILE, 
    SH_ERR_T_NAME        
  };
    

  /* interpret line                                    */

  c = strchr(line, '=');
  ++c;

  if (c == NULL)         /* no 'key=value' pair        */
    return;

  while ((*c) == ' ' || (*c) == '\t')
    ++c;

  if ((*c) == '\0')     /* no value                    */
    return;

  /* clip off trailing white space                     */

  tmp = line + sl_strlen( line ); --tmp;
  while( isspace( (int) *tmp ) && tmp >= line ) 
    *tmp-- = '\0';

  if (!sl_is_suid())
    sh_error_handle (SH_ERR_ALL,
		     _("ConfigFile: %s"), 
		     line);

  /* file or directory                                 */

  if      (sl_strncmp (line, _("dir"),  sizeof("dir")-1)  == 0) 
    key = 2;
  else if (sl_strncmp (line, _("file"), sizeof("file")-1) == 0) 
    key = 1;
  else
    key = 0;

#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
    
  if      (read_mode == SH_SECTION_ATTRIBUTES) {
    if      (key == 1) sh_files_pushfile_attr (c);
    else if (key == 2) sh_files_pushdir_attr  (c);
  }
  else if      (read_mode == SH_SECTION_READONLY) {
    if      (key == 1) sh_files_pushfile_ro (c);
    else if (key == 2) sh_files_pushdir_ro  (c);
  }
  else if (read_mode == SH_SECTION_LOGFILES) {
    if      (key == 1) sh_files_pushfile_log (c);
    else if (key == 2) sh_files_pushdir_log  (c);
  }
  else if (read_mode == SH_SECTION_LOGGROW) {
    if      (key == 1) sh_files_pushfile_glog (c);
    else if (key == 2) sh_files_pushdir_glog  (c);
  }
  else if (read_mode == SH_SECTION_NOIGNORE) {
    if      (key == 1) sh_files_pushfile_noig (c);
    else if (key == 2) sh_files_pushdir_noig  (c);
  }
  else if (read_mode == SH_SECTION_ALLIGNORE) {
    if      (key == 1) sh_files_pushfile_allig (c);
    else if (key == 2) sh_files_pushdir_allig  (c);
  }

  else if (read_mode == SH_SECTION_MISC) {
#else
  if (read_mode == SH_SECTION_MISC) {
#endif
    if (sl_strncmp (line,  _("SetFilecheckTime"), 
		 sizeof  ("SetFilecheckTime")-1) == 0) {
      sh_readconf_setTime (c, SET_FILETIME);
    }
#ifdef SH_WITH_SERVER
    if (sl_strncmp (line,  _("SetClientTimeLimit"), 
		 sizeof  ("SetClientTimeLimit")-1) == 0) {
      sh_forward_set_time_limit (c);
    }
#endif
    else if (sl_strncmp (line,  _("SetMailTime"), 
		      sizeof  ("SetMailTime")-1) == 0) {
      sh_readconf_setTime (c, SET_MAILTIME);
    }
    else if (sl_strncmp (line,  _("SetMailNum"), 
		      sizeof  ("SetMailNum")-1) == 0) {
      sh_mail_setNum (c);
    }
    else if (sl_strncmp (line,  _("SetMailAddress"), 
		      sizeof  ("SetMailAddress")-1) == 0) {
      sh_mail_setaddress (c);
    }
    else if (sl_strncmp (line,  _("SetMailRelay"), 
		      sizeof  ("SetMailRelay")-1) == 0) {
      sh_mail_set_relay (c);
    }
    else if (sl_strncmp (line,  _("SamhainPath"), 
		      sizeof  ("SamhainPath")-1) == 0) {
      sh_unix_self_hash (c);
    }
    else if (sl_strncmp (line,  _("Daemon"), 
		      sizeof  ("Daemon")-1) == 0) {
      sh_unix_setdeamon(c);
    }
    else if (sl_strncmp (line,  _("TrustedUser"), 
		      sizeof  ("TrustedUser")-1) == 0) {
      tf_add_trusted_user (c);
    }
    else if (sl_strncmp (line,  _("SetTimeServer"), 
		      sizeof  ("SetTimeServer")-1) == 0) {
      sh_unix_settimeserver (c);
    }
#ifdef SH_WITH_CLIENT
    else if (sl_strncmp (line,  _("SetLogServer"), 
		      sizeof("SetLogServer")-1) == 0) {
      sh_forward_setlogserver (c);
    }
#endif
    else if (sl_strncmp (line,  _("SetLoopTime"), 
		      sizeof("SetLoopTime")-1) == 0) {
      sh_util_setlooptime (c);
    }
#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
    else if (sl_strncmp (line,  _("SetRecursionLevel"), 
		      sizeof("SetRecursionLevel")-1) == 0) {
      sh_files_setrecursion (c);
    }
    else if (sl_strncmp (line,  _("ChecksumTest"), 
		      sizeof("ChecksumTest")-1) == 0) {
      sh_util_setchecksum (c);
    }
#endif
  }

#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
  else if      (read_mode == SH_SECTION_OTHER) {
    for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
      {
	for (modkey = 0; modList[modnum].conf_table[modkey].the_opt != NULL; 
	     ++modkey) 
	  {
	    if (sl_strncmp (line,  _(modList[modnum].conf_table[modkey].the_opt),
		     sl_strlen(modList[modnum].conf_table[modkey].the_opt)) == 0)
	      modList[modnum].conf_table[modkey].func(c);
	  }
      }
  }
#endif


#ifdef SH_WITH_SERVER
  else if (read_mode == SH_SECTION_CLIENTS) {
    if      (sl_strncmp (line,  _("Client"), 
		      sizeof("Client")-1) == 0) 
      sh_forward_register_client (c);
  }
#endif

  else if (read_mode == SH_SECTION_LOG) {
    if (sl_strncmp (line,  _("MailSeverity"), 
		      sizeof("MailSeverity")-1) == 0)
      sh_error_setseverity (c);
    else if (sl_strncmp (line,  _("PrintSeverity"), 
		      sizeof("PrintSeverity")-1) == 0)
      sh_error_setprint (c);
#ifdef SH_WITH_CLIENT
    else if (sl_strncmp (line,  _("ExportSeverity"), 
		      sizeof("ExportSeverity")-1) == 0)
      sh_error_setexport (c);
#endif
#ifdef WITH_EXTERNAL
    else if (sl_strncmp (line,  _("ExternalSeverity"), 
		      sizeof("ExternalSeverity")-1) == 0)
      sh_error_set_external (c);
#endif
    else if (sl_strncmp (line,  _("LogSeverity"), 
		      sizeof("LogSeverity")-1) == 0)
      sh_error_setlog (c);
    else if (sl_strncmp (line,  _("SyslogSeverity"), 
		      sizeof("SyslogSeverity")-1) == 0)
      sh_error_set_syslog (c);
  }
  
  else if (read_mode == SH_SECTION_THRESHOLD) {
    i = 0;
    while (ident[i] != NULL) {
      if (sl_strncmp (line, _(ident[i]), sl_strlen(ident[i])-1) == 0)
	sh_error_set_iv (identnum[i], c);
      ++i;
    }
  }

#ifdef WITH_EXTERNAL
  else if (read_mode == SH_SECTION_EXTERNAL) 
    {
      i = 0;
      while (ext_table[i].optname != NULL) 
	{
	  if (sl_strncmp (line, _(ext_table[i].optname), 
			  sl_strlen(ext_table[i].optname)-1) == 0)
	    {
	      if (0 != ext_table[i].func (c))
		sh_error_handle (SH_ERR_ERR,
				 _("Config External: %s: %s"),
				 _(ext_table[i].optname), c);
;
	      break;
	    }
	  ++i;
	}
    }
#endif

     
  return;
}
  
    
